
/* Copyright (C) 2015 David Eller <david@larosterna.com>
 * 
 * Commercial License Usage
 * Licensees holding valid commercial licenses may use this file in accordance
 * with the terms contained in their respective non-exclusive license agreement.
 * For further information contact david@larosterna.com .
 *
 * GNU General Public License Usage
 * Alternatively, this file may be used under the terms of the GNU General
 * Public License version 3.0 as published by the Free Software Foundation and
 * appearing in the file gpl.txt included in the packaging of this file.
 */
 
#include "volwavedrag.h"
#include "trimesh.h"
#include "mxmesh.h"

#include <iostream>
using namespace std;

// precomputed coeffients for sine series with 48 terms below
const int Nc = 48;

VolWaveDrag::VolWaveDrag(const TriMesh &tm)
{
  const int nf = tm.nfaces();
  vtri.resize(3*nf);
  const PointList<3> & vtx( tm.vertices() );
  for (int i=0; i<nf; ++i) {
    const uint *vi = tm.face(i).vertices();
    for (int k=0; k<3; ++k)
      vtri[3*i+k] = vtx[ vi[k] ];
  }
  initBox();
}

VolWaveDrag::VolWaveDrag(const MxMesh &mx)
{
  const PointList<3> & vtx( mx.nodes() );
  const int nsec = mx.nsections();
  for (int j=0; j<nsec; ++j) {
    const MxMeshSection & sec( mx.section(j) );
    if (sec.elementType() == Mx::Tri3) {
      const int ne = sec.nelements();
      PointList<3> tmp(3*ne);
      for (int i=0; i<ne; ++i) {
        const uint *vi = sec.element(i);
        for (int k=0; k<3; ++k)
          tmp[3*i+k] = vtx[vi[k]];
      }
      vtri.insert(vtri.end(), tmp.begin(), tmp.end());
    }
  }
  initBox();
}

void VolWaveDrag::initBox()
{
  bbp1 = std::numeric_limits<Real>::max();
  bbp2 = -bbp1;
  const int np = vtri.size();
  for (int i=0; i<np; ++i) {
    const Vct3 & p( vtri[i] );
    for (int k=0; k<3; ++k) {
      bbp1[k] = std::min( bbp1[k], p[k] );
      bbp2[k] = std::max( bbp2[k], p[k] );
    }
  }
}

void VolWaveDrag::meanAreaDistribution(const Vct3 &pn, Real Mach, int nsec,
                                       int nphi, Vector &x, Vector &S) const
{
  // reference sections for M = 1
  areaDistribution(pn, nsec, x, S);
  if (Mach == 1.0)
    return;

  // create two vectors which are perpendicular to pn and each other
  Vct3 xn(pn), yax, zax;
  extend_basis(xn, yax, zax);

  // mach cone angle sin(mu) = 1/M
  Real smu, cmu;
  smu = 1.0 / Mach;
  cmu = sqrt(1.0 - sq(smu));

  S = 0.0;
  Real wgt = smu / nphi;

  // parallelization benefits from using just the physical cores and
  // not the virtual ones;

#pragma omp parallel num_threads(4)
  {
    Vector xt, St, Sm(nsec);

#pragma omp for
    for (int j=0; j<nphi; ++j) {
      Real sphi, cphi, phi = 2*PI / (nphi-1);
      sincosine(phi, sphi, cphi);
      Vct3 pjn = smu*xn + cmu*(sphi*yax + cphi*zax);
      areaDistribution(pjn, nsec, xt, St);
      Sm += wgt*St;
    }

#pragma omp critical
    S += Sm;
  }
}

void VolWaveDrag::areaDistribution(const Vct3 &pn, int n,
                                   Vector &x, Vector &S) const
{
  // compute distance of all triangle vertices from origin
  // in the direction of the plane normal
  Vector dst;
  Vct3 pnn(pn);
  Real dmin, dmax;
  normalize(pnn);
  planeDistance(pnn, dst, dmin, dmax);

  x.allocate(n);
  S.allocate(n);

  Indices itri;
  for (int i=0; i<n; ++i) {
    Real t = Real(i) / (n-1);
    x[i] = dmin + t*(dmax - dmin);
    intersectingTriangles(dst, x[i], itri);
    const int nt = itri.size();
    S[i] = 0.0;
    for (int j=0; j<nt; ++j)
      S[i] += areaContribution(pnn, x[i], itri[j]);
  }
}

Real VolWaveDrag::areaContribution(const Vct3 &pn, Real dp, uint jt) const
{
  // determine edge line parameter where edge intersects plane
  Vct3 isp[3];
  const Vct3 *p = &vtri[3*jt];
  int ip = 0;
  for (int k=0; k<3; ++k) {
    const Vct3 & src( p[k] );
    const Vct3 & trg( p[(k+1)%3] );
    Real xa = dot(pn, src);
    Real xb = dot(pn, trg);
    if (xb == xa)
      continue;
    Real t = (dp - xa) / (xb - xa);
    if (t >= 0.0 and t <= 1.0) {
      isp[ip] = (1 - t)*src + t*trg;
      ++ip;
    }
  }

  // triangle normal
  Vct3 fn = cross( p[1]-p[0], p[2]-p[0] );
  if (ip == 2) {

    // origin of the integration, a point in the plane
    Vct3 pctr = dp*pn;

    // positive or negative contribution ?
    Vct3 etc = 0.5*(isp[0] + isp[1]) - pctr;
    Real s = sign( dot(fn, etc) );

    // scaled normal of the triangle from pctr to isp[0] and isp[1]
    Real snt = 0.5 * norm( cross( isp[0]-pctr, isp[1]-pctr ) );

    // area of this triangle, projected onto plane pn
    // this assumes pn is normalized
    return s * snt;
  }

  // only one intersection : point touches plane, no area contribution
  if (ip < 2)
    return 0.0;

  // final possibility : triangle is exactly in plane
  // count entire triangle area as contribution
  return 0.5*norm(fn);
}

void VolWaveDrag::planeDistance(const Vct3 & pn, Vector &dst,
                                Real & dmin, Real & dmax) const
{
  const int nv = vtri.size();
  dst.allocate(nv);
  dmin = std::numeric_limits<Real>::max();
  dmax = -dmin;

  for (int i=0; i<nv; ++i) {
    dst[i] = dot(pn, vtri[i]);
    dmin = std::min(dmin, dst[i]);
    dmax = std::max(dmax, dst[i]);
  }
}

void VolWaveDrag::intersectingTriangles(const Vector &dst, Real dp,
                                        Indices &itri) const
{
  itri.clear();
  const int ntri = vtri.size() / 3;
  for (int i=0; i<ntri; ++i) {
    int nbeyond = 0;
    for (int k=0; k<3; ++k)
      nbeyond += (dst[3*i+k] >= dp);
    if (nbeyond > 0 and nbeyond < 3)
      itri.push_back(i);
  }
}

Real VolWaveDrag::dragCoefficient(Real Sref, const Vector &x,
                                  const Vector &S, Real Scutoff)
{
  assert(x.size() == S.size());

  // peel off values to the left/right
  Real Smax = max(S);
  const Real peelS = Scutoff*Smax;
  const int np = S.size();
  int ix1(0), ix2(np);
  while ( ix1 < np and S[ix1] < peelS )
    ++ix1;
  while ( ix2 > 0 and S[ix2-1] < peelS )
    --ix2;

  //  if (S[ix1] > 0)
  //    --ix1;
  //  if (ix2 < np and S[ix2-1] > 0)
  //    ++ix2;

  if (ix2 <= ix1)
    return 0.0;

  Real L = x[ix2-1] - x[ix1];
  Real S1 = S[ix1];    // nose area
  Real S2 = S[ix2-1];  // base area

  // cout << "L " << L << " S1 " << S1 << " S2 " << S2 << endl;

  const int nx = ix2 - ix1;
  Vector eta(nx), Se(nx);
  for (int i=0; i<nx; ++i) {
    eta[i] = (x[ix1+i] - x[ix1]) / L;
    Se[i] = S[ix1+i] - (S1 + (S2 - S1)*eta[i]);
  }

  //  cout << "eta, area" << endl;
  //  for (int i=0; i<nx; ++i)
  //    cout << i+1 << " " << eta[i] << ", " << Se[i] << endl;

  Vector ak(Nc);
  sineSeries(eta, Se, ak);

  // cout << "ak " << ak << endl;

  // TODO:
  // implement drag contribution from base/nose areas

  return dragCoefficient(Sref, L, ak);
}

void VolWaveDrag::sineSeries(const Vector &eta, const Vector &S,
                             Vector &ak)
{
  // 2 * \int S(\eta) * sin(pi*k*\eta) d \eta
  ak.allocate(Nc);
  const int np = eta.size();
  for (int k=0; k<Nc; ++k) {
    ak[k] = 0.0;
    for (int i=1; i<np; ++i) {
      Real dx = eta[i] - eta[i-1];
      Real S1 = S[i-1] * std::sin( M_PI*(k+1)*eta[i-1] );
      Real S2 = S[i] * std::sin( M_PI*(k+1)*eta[i] );
      ak[k] += (S1 + S2)*dx;
    }
  }
}


static const double A48[48 * 48] = {
  -7.021816085628420e+01, 0.000000000000000e+00, -8.900824650135900e+01, 0.000000000000000e+00, -1.557732790433592e+02,
  0.000000000000000e+00, -2.219421338067887e+02, 0.000000000000000e+00, -2.877811496236712e+02, 0.000000000000000e+00,
  -3.534220607884801e+02, 0.000000000000000e+00, -4.189346438947925e+02, 0.000000000000000e+00, -4.843591034511386e+02,
  0.000000000000000e+00, -5.497202556151707e+02, 0.000000000000000e+00, -6.150342792154646e+02, 0.000000000000000e+00,
  -6.803121949128343e+02, 0.000000000000000e+00, -7.455617826369836e+02, 0.000000000000000e+00, -8.107886993610502e+02,
  0.000000000000000e+00, -8.759971625273406e+02, 0.000000000000000e+00, -9.411903851588594e+02, 0.000000000000000e+00,
  -1.006370862428824e+03, 0.000000000000000e+00, -1.071540565398104e+03, 0.000000000000000e+00, -1.136701073987475e+03,
  0.000000000000000e+00, -1.201853672227242e+03, 0.000000000000000e+00, -1.266999411449095e+03, 0.000000000000000e+00,
  -1.332139160330905e+03, 0.000000000000000e+00, -1.397273639967767e+03, 0.000000000000000e+00, -1.462403456704839e+03,
  0.000000000000000e+00, -1.527529121258250e+03, 0.000000000000000e+00, 0.000000000000000e+00, -5.442422038731835e+02,
  0.000000000000000e+00, -3.493185644110063e+02, 0.000000000000000e+00, -5.454656144657413e+02, 0.000000000000000e+00,
  -7.411309110801784e+02, 0.000000000000000e+00, -9.362273780629544e+02, 0.000000000000000e+00, -1.130859948859139e+03,
  0.000000000000000e+00, -1.325133358413047e+03, 0.000000000000000e+00, -1.519128395711831e+03, 0.000000000000000e+00,
  -1.712905067652055e+03, 0.000000000000000e+00, -1.906507802644545e+03, 0.000000000000000e+00, -2.099969957849160e+03,
  0.000000000000000e+00, -2.293316673891452e+03, 0.000000000000000e+00, -2.486567997506928e+03, 0.000000000000000e+00,
  -2.679738577839147e+03, 0.000000000000000e+00, -2.872841567736994e+03, 0.000000000000000e+00, -3.065885822368984e+03,
  0.000000000000000e+00, -3.258879984613314e+03, 0.000000000000000e+00, -3.451830367885173e+03, 0.000000000000000e+00,
  -3.644742275450894e+03, 0.000000000000000e+00, -3.837620686586279e+03, 0.000000000000000e+00, -4.030469242118391e+03,
  0.000000000000000e+00, -4.223291037012832e+03, 0.000000000000000e+00, -4.416089645564901e+03, 0.000000000000000e+00,
  -4.608866029704063e+03, -8.900824650135900e+01, 0.000000000000000e+00, -1.723600100944841e+03, 0.000000000000000e+00,
  -7.458023379325588e+02, 0.000000000000000e+00, -1.084959297976301e+03, 0.000000000000000e+00, -1.423766507479137e+03,
  0.000000000000000e+00, -1.761854016055015e+03, 0.000000000000000e+00, -2.099241198775179e+03, 0.000000000000000e+00,
  -2.436024343584000e+03, 0.000000000000000e+00, -2.772302389555130e+03, 0.000000000000000e+00, -3.108161365902908e+03,
  0.000000000000000e+00, -3.443671281577793e+03, 0.000000000000000e+00, -3.778888613051527e+03, 0.000000000000000e+00,
  -4.113858728504948e+03, 0.000000000000000e+00, -4.448618762383861e+03, 0.000000000000000e+00, -4.783197432539635e+03,
  0.000000000000000e+00, -5.117620385308717e+03, 0.000000000000000e+00, -5.451908258204180e+03, 0.000000000000000e+00,
  -5.786076011711925e+03, 0.000000000000000e+00, -6.120139641823457e+03, 0.000000000000000e+00, -6.454110686901437e+03,
  0.000000000000000e+00, -6.787998439043877e+03, 0.000000000000000e+00, -7.121813205198619e+03, 0.000000000000000e+00,
  -7.455561197379555e+03, 0.000000000000000e+00, -7.789249104757998e+03, 0.000000000000000e+00, 0.000000000000000e+00,
  -3.493185644110063e+02, 0.000000000000000e+00, -3.944646287887625e+03, 0.000000000000000e+00, -1.322998661509320e+03,
  0.000000000000000e+00, -1.822212201698135e+03, 0.000000000000000e+00, -2.321961156781218e+03, 0.000000000000000e+00,
  -2.821324171328721e+03, 0.000000000000000e+00, -3.320053475538520e+03, 0.000000000000000e+00, -3.818122992841325e+03,
  0.000000000000000e+00, -4.315578064443836e+03, 0.000000000000000e+00, -4.812484119847105e+03, 0.000000000000000e+00,
  -5.308906194856751e+03, 0.000000000000000e+00, -5.804906348815684e+03, 0.000000000000000e+00, -6.300536986920059e+03,
  0.000000000000000e+00, -6.795843707256664e+03, 0.000000000000000e+00, -7.290865619606987e+03, 0.000000000000000e+00,
  -7.785636442245068e+03, 0.000000000000000e+00, -8.280185354358528e+03, 0.000000000000000e+00, -8.774536985201288e+03,
  0.000000000000000e+00, -9.268710549823660e+03, 0.000000000000000e+00, -9.762725985126383e+03, 0.000000000000000e+00,
  -1.025659846127254e+04, 0.000000000000000e+00, -1.075034369139816e+04, 0.000000000000000e+00, -1.124396941732769e+04,
  0.000000000000000e+00, -1.173749012291644e+04, -1.557732790433592e+02, 0.000000000000000e+00, -7.458023379325588e+02,
  0.000000000000000e+00, -7.481432730704668e+03, 0.000000000000000e+00, -2.055298360510070e+03, 0.000000000000000e+00,
  -2.721467464193927e+03, 0.000000000000000e+00, -3.388638002567237e+03, 0.000000000000000e+00, -4.055653398689695e+03,
  0.000000000000000e+00, -4.722098055600878e+03, 0.000000000000000e+00, -5.387843860564923e+03, 0.000000000000000e+00,
  -6.052884380854417e+03, 0.000000000000000e+00, -6.717258033018387e+03, 0.000000000000000e+00, -7.381020454526882e+03,
  0.000000000000000e+00, -8.044231668697851e+03, 0.000000000000000e+00, -8.706945752089758e+03, 0.000000000000000e+00,
  -9.369216356199058e+03, 0.000000000000000e+00, -1.003108929665355e+04, 0.000000000000000e+00, -1.069260622725354e+04,
  0.000000000000000e+00, -1.135380234444897e+04, 0.000000000000000e+00, -1.201470818596153e+04, 0.000000000000000e+00,
  -1.267535271733391e+04, 0.000000000000000e+00, -1.333576218715648e+04, 0.000000000000000e+00, -1.399595621755738e+04,
  0.000000000000000e+00, -1.465595513085652e+04, 0.000000000000000e+00, -1.531577253265984e+04, 0.000000000000000e+00,
  0.000000000000000e+00, -5.454656144657413e+02, 0.000000000000000e+00, -1.322998661509320e+03, 0.000000000000000e+00,
  -1.266566091820405e+04, 0.000000000000000e+00, -2.982807372929191e+03, 0.000000000000000e+00, -3.825608168264319e+03,
  0.000000000000000e+00, -4.670301890999775e+03, 0.000000000000000e+00, -5.515381923525639e+03, 0.000000000000000e+00,
  -6.360188231560751e+03, 0.000000000000000e+00, -7.204439940935847e+03, 0.000000000000000e+00, -8.048028238161050e+03,
  0.000000000000000e+00, -8.890936283542085e+03, 0.000000000000000e+00, -9.733182233855274e+03, 0.000000000000000e+00,
  -1.057480512673283e+04, 0.000000000000000e+00, -1.141584891033131e+04, 0.000000000000000e+00, -1.225636298288041e+04,
  0.000000000000000e+00, -1.309638705352119e+04, 0.000000000000000e+00, -1.393596630073454e+04, 0.000000000000000e+00,
  -1.477514032106358e+04, 0.000000000000000e+00, -1.561394176606279e+04, 0.000000000000000e+00, -1.645240485431484e+04,
  0.000000000000000e+00, -1.729055685911823e+04, 0.000000000000000e+00, -1.812841989874500e+04, 0.000000000000000e+00,
  -1.896602166840895e+04, 0.000000000000000e+00, -1.980338156304536e+04, -2.219421338067887e+02, 0.000000000000000e+00,
  -1.084959297976301e+03, 0.000000000000000e+00, -2.055298360510070e+03, 0.000000000000000e+00, -1.976574746360253e+04,
  0.000000000000000e+00, -4.078612015520396e+03, 0.000000000000000e+00, -5.102958547938114e+03, 0.000000000000000e+00,
  -6.129670936044879e+03, 0.000000000000000e+00, -7.157123372665912e+03, 0.000000000000000e+00, -8.184523133534467e+03,
  0.000000000000000e+00, -9.211482516356036e+03, 0.000000000000000e+00, -1.023782368359479e+04, 0.000000000000000e+00,
  -1.126347034788494e+04, 0.000000000000000e+00, -1.228841447024543e+04, 0.000000000000000e+00, -1.331267141125037e+04,
  0.000000000000000e+00, -1.433627404487471e+04, 0.000000000000000e+00, -1.535926136292369e+04, 0.000000000000000e+00,
  -1.638167347812988e+04, 0.000000000000000e+00, -1.740355714564152e+04, 0.000000000000000e+00, -1.842494263862793e+04,
  0.000000000000000e+00, -1.944587313277898e+04, 0.000000000000000e+00, -2.046638393962805e+04, 0.000000000000000e+00,
  -2.148650379292331e+04, 0.000000000000000e+00, -2.250626287401369e+04, 0.000000000000000e+00, -2.352568895009513e+04,
  0.000000000000000e+00, 0.000000000000000e+00, -7.411309110801784e+02, 0.000000000000000e+00, -1.822212201698135e+03,
  0.000000000000000e+00, -2.982807372929191e+03, 0.000000000000000e+00, -2.911134633260641e+04, 0.000000000000000e+00,
  -5.380519950602084e+03, 0.000000000000000e+00, -6.593106539329879e+03, 0.000000000000000e+00, -7.808819101420856e+03,
  0.000000000000000e+00, -9.025833994614293e+03, 0.000000000000000e+00, -1.024318449115097e+04, 0.000000000000000e+00,
  -1.146034647150601e+04, 0.000000000000000e+00, -1.267704254574228e+04, 0.000000000000000e+00, -1.389313330238267e+04,
  0.000000000000000e+00, -1.510854699002579e+04, 0.000000000000000e+00, -1.632327568583192e+04, 0.000000000000000e+00,
  -1.753732390748895e+04, 0.000000000000000e+00, -1.875071407578987e+04, 0.000000000000000e+00, -1.996347869006161e+04,
  0.000000000000000e+00, -2.117565016302377e+04, 0.000000000000000e+00, -2.238726351686981e+04, 0.000000000000000e+00,
  -2.359835989651236e+04, 0.000000000000000e+00, -2.480895994525463e+04, 0.000000000000000e+00, -2.601910769426565e+04,
  0.000000000000000e+00, -2.722882647230378e+04, 0.000000000000000e+00, -2.843814955052699e+04, -2.877811496236712e+02,
  0.000000000000000e+00, -1.423766507479137e+03, 0.000000000000000e+00, -2.721467464193927e+03, 0.000000000000000e+00,
  -4.078612015520396e+03, 0.000000000000000e+00, -4.096848682224598e+04, 0.000000000000000e+00, -6.860629882692409e+03,
  0.000000000000000e+00, -8.265335600249755e+03, 0.000000000000000e+00, -9.673580931147397e+03, 0.000000000000000e+00,
  -1.108349840867668e+04, 0.000000000000000e+00, -1.249403222670351e+04, 0.000000000000000e+00, -1.390457634887765e+04,
  0.000000000000000e+00, -1.531478093423432e+04, 0.000000000000000e+00, -1.672444964297231e+04, 0.000000000000000e+00,
  -1.813348257048470e+04, 0.000000000000000e+00, -1.954181899378723e+04, 0.000000000000000e+00, -2.094945732202523e+04,
  0.000000000000000e+00, -2.235640155059753e+04, 0.000000000000000e+00, -2.376266941878928e+04, 0.000000000000000e+00,
  -2.516828985162449e+04, 0.000000000000000e+00, -2.657329120666107e+04, 0.000000000000000e+00, -2.797770788827053e+04,
  0.000000000000000e+00, -2.938156716047105e+04, 0.000000000000000e+00, -3.078491197016141e+04, 0.000000000000000e+00,
  -3.218775443098358e+04, 0.000000000000000e+00, 0.000000000000000e+00, -9.362273780629544e+02, 0.000000000000000e+00,
  -2.321961156781218e+03, 0.000000000000000e+00, -3.825608168264319e+03, 0.000000000000000e+00, -5.380519950602084e+03,
  0.000000000000000e+00, -5.566565566629184e+04, 0.000000000000000e+00, -8.555448360795028e+03, 0.000000000000000e+00,
  -1.015736574801152e+04, 0.000000000000000e+00, -1.176345046426734e+04, 0.000000000000000e+00, -1.337173102193114e+04,
  0.000000000000000e+00, -1.498103270150460e+04, 0.000000000000000e+00, -1.659064676188475e+04, 0.000000000000000e+00,
  -1.820013706400732e+04, 0.000000000000000e+00, -1.980923929204588e+04, 0.000000000000000e+00, -2.141779889431290e+04,
  0.000000000000000e+00, -2.302573220300199e+04, 0.000000000000000e+00, -2.463297826597230e+04, 0.000000000000000e+00,
  -2.623953905598289e+04, 0.000000000000000e+00, -2.784541112097300e+04, 0.000000000000000e+00, -2.945060909629598e+04,
  0.000000000000000e+00, -3.105515224699034e+04, 0.000000000000000e+00, -3.265906906650026e+04, 0.000000000000000e+00,
  -3.426238446975283e+04, 0.000000000000000e+00, -3.586512709803659e+04, 0.000000000000000e+00, -3.746732406705313e+04,
  -3.534220607884801e+02, 0.000000000000000e+00, -1.761854016055015e+03, 0.000000000000000e+00, -3.388638002567237e+03,
  0.000000000000000e+00, -5.102958547938114e+03, 0.000000000000000e+00, -6.860629882692409e+03, 0.000000000000000e+00,
  -7.346756906896211e+04, 0.000000000000000e+00, -1.043638757939430e+04, 0.000000000000000e+00, -1.223873824871137e+04,
  0.000000000000000e+00, -1.404560620237454e+04, 0.000000000000000e+00, -1.585501716657124e+04, 0.000000000000000e+00,
  -1.766574600903898e+04, 0.000000000000000e+00, -1.947701802402329e+04, 0.000000000000000e+00, -2.128834434914897e+04,
  0.000000000000000e+00, -2.309940740131831e+04, 0.000000000000000e+00, -2.491001115513829e+04, 0.000000000000000e+00,
  -2.672003474298904e+04, 0.000000000000000e+00, -2.852941452646049e+04, 0.000000000000000e+00, -3.033809183099653e+04,
  0.000000000000000e+00, -3.214607770814979e+04, 0.000000000000000e+00, -3.395336555015762e+04, 0.000000000000000e+00,
  -3.575996760516313e+04, 0.000000000000000e+00, -3.756590081506492e+04, 0.000000000000000e+00, -3.937119190166871e+04,
  0.000000000000000e+00, -4.117586089622098e+04, 0.000000000000000e+00, 0.000000000000000e+00, -1.130859948859139e+03,
  0.000000000000000e+00, -2.821324171328721e+03, 0.000000000000000e+00, -4.670301890999775e+03, 0.000000000000000e+00,
  -6.593106539329879e+03, 0.000000000000000e+00, -8.555448360795028e+03, 0.000000000000000e+00, -9.470195910226794e+04,
  0.000000000000000e+00, -1.253912474808586e+04, 0.000000000000000e+00, -1.454600650712510e+04, 0.000000000000000e+00,
  -1.655792346161379e+04, 0.000000000000000e+00, -1.857285239980273e+04, 0.000000000000000e+00, -2.058949154564191e+04,
  0.000000000000000e+00, -2.260698818343645e+04, 0.000000000000000e+00, -2.462478625987730e+04, 0.000000000000000e+00,
  -2.664250853429418e+04, 0.000000000000000e+00, -2.865991018748078e+04, 0.000000000000000e+00, -3.067683003429458e+04,
  0.000000000000000e+00, -3.269316790123957e+04, 0.000000000000000e+00, -3.470887072242604e+04, 0.000000000000000e+00,
  -3.672387439603401e+04, 0.000000000000000e+00, -3.873819661472266e+04, 0.000000000000000e+00, -4.075182608300554e+04,
  0.000000000000000e+00, -4.276477248623956e+04, 0.000000000000000e+00, -4.477704804260290e+04, 0.000000000000000e+00,
  -4.678867343710619e+04, -4.189346438947925e+02, 0.000000000000000e+00, -2.099241198775179e+03, 0.000000000000000e+00,
  -4.055653398689695e+03, 0.000000000000000e+00, -6.129670936044879e+03, 0.000000000000000e+00, -8.265335600249755e+03,
  0.000000000000000e+00, -1.043638757939430e+04, 0.000000000000000e+00, -1.196327122676822e+05, 0.000000000000000e+00,
  -1.483457744474186e+04, 0.000000000000000e+00, -1.704875056093422e+04, 0.000000000000000e+00, -1.926825374289000e+04,
  0.000000000000000e+00, -2.149108363413323e+04, 0.000000000000000e+00, -2.371591060733285e+04, 0.000000000000000e+00,
  -2.594184174914374e+04, 0.000000000000000e+00, -2.816826899311096e+04, 0.000000000000000e+00, -3.039478142617566e+04,
  0.000000000000000e+00, -3.262109138599801e+04, 0.000000000000000e+00, -3.484700783249143e+04, 0.000000000000000e+00,
  -3.707240069370350e+04, 0.000000000000000e+00, -3.929718873955939e+04, 0.000000000000000e+00, -4.152133388419947e+04,
  0.000000000000000e+00, -4.374476465282483e+04, 0.000000000000000e+00, -4.596751016628496e+04, 0.000000000000000e+00,
  -4.818955733501816e+04, 0.000000000000000e+00, -5.041091321403611e+04, 0.000000000000000e+00, 0.000000000000000e+00,
  -1.325133358413047e+03, 0.000000000000000e+00, -3.320053475538520e+03, 0.000000000000000e+00, -5.515381923525639e+03,
  0.000000000000000e+00, -7.808819101420856e+03, 0.000000000000000e+00, -1.015736574801152e+04, 0.000000000000000e+00,
  -1.253912474808586e+04, 0.000000000000000e+00, -1.485870334217491e+05, 0.000000000000000e+00, -1.735784890399001e+04,
  0.000000000000000e+00, -1.978273523200483e+04, 0.000000000000000e+00, -2.221338355931084e+04, 0.000000000000000e+00,
  -2.464777235380881e+04, 0.000000000000000e+00, -2.708452528261843e+04, 0.000000000000000e+00, -2.952269358841811e+04,
  0.000000000000000e+00, -3.196161317596226e+04, 0.000000000000000e+00, -3.440082799216398e+04, 0.000000000000000e+00,
  -3.684000511242408e+04, 0.000000000000000e+00, -3.927891667675359e+04, 0.000000000000000e+00, -4.171740431337220e+04,
  0.000000000000000e+00, -4.415535676459959e+04, 0.000000000000000e+00, -4.659270496871221e+04, 0.000000000000000e+00,
  -4.902941583009261e+04, 0.000000000000000e+00, -5.146541270251458e+04, 0.000000000000000e+00, -5.390073605170132e+04,
  0.000000000000000e+00, -5.633536375283465e+04, -4.843591034511386e+02, 0.000000000000000e+00, -2.436024343584000e+03,
  0.000000000000000e+00, -4.722098055600878e+03, 0.000000000000000e+00, -7.157123372665912e+03, 0.000000000000000e+00,
  -9.673580931147397e+03, 0.000000000000000e+00, -1.223873824871137e+04, 0.000000000000000e+00, -1.483457744474186e+04,
  0.000000000000000e+00, -1.818282372648472e+05, 0.000000000000000e+00, -2.007948592176624e+04, 0.000000000000000e+00,
  -2.271750262761901e+04, 0.000000000000000e+00, -2.536153099287273e+04, 0.000000000000000e+00, -2.800958061992193e+04,
  0.000000000000000e+00, -3.066026488679739e+04, 0.000000000000000e+00, -3.331260690566686e+04, 0.000000000000000e+00,
  -3.596591041251293e+04, 0.000000000000000e+00, -3.861967581614530e+04, 0.000000000000000e+00, -4.127355149244117e+04,
  0.000000000000000e+00, -4.392727276020605e+04, 0.000000000000000e+00, -4.658065548237089e+04, 0.000000000000000e+00,
  -4.923356999207695e+04, 0.000000000000000e+00, -5.188592439452296e+04, 0.000000000000000e+00, -5.453765658536724e+04,
  0.000000000000000e+00, -5.718875255823218e+04, 0.000000000000000e+00, -5.983911452224748e+04, 0.000000000000000e+00,
  0.000000000000000e+00, -1.519128395711831e+03, 0.000000000000000e+00, -3.818122992841325e+03, 0.000000000000000e+00,
  -6.360188231560751e+03, 0.000000000000000e+00, -9.025833994614293e+03, 0.000000000000000e+00, -1.176345046426734e+04,
  0.000000000000000e+00, -1.454600650712510e+04, 0.000000000000000e+00, -1.735784890399001e+04, 0.000000000000000e+00,
  -2.196831388622906e+05, 0.000000000000000e+00, -2.303417126438240e+04, 0.000000000000000e+00, -2.588825260594868e+04,
  0.000000000000000e+00, -2.874870938689374e+04, 0.000000000000000e+00, -3.161355207786552e+04, 0.000000000000000e+00,
  -3.448136609182136e+04, 0.000000000000000e+00, -3.735113609091195e+04, 0.000000000000000e+00, -4.022212463470113e+04,
  0.000000000000000e+00, -4.309378901743732e+04, 0.000000000000000e+00, -4.596574782348224e+04, 0.000000000000000e+00,
  -4.883769700288491e+04, 0.000000000000000e+00, -5.170942695435766e+04, 0.000000000000000e+00, -5.458078395316319e+04,
  0.000000000000000e+00, -5.745165428250096e+04, 0.000000000000000e+00, -6.032195726519755e+04, 0.000000000000000e+00,
  -6.319164134422907e+04, 0.000000000000000e+00, -6.606069443269838e+04, -5.497202556151707e+02, 0.000000000000000e+00,
  -2.772302389555130e+03, 0.000000000000000e+00, -5.387843860564923e+03, 0.000000000000000e+00, -8.184523133534467e+03,
  0.000000000000000e+00, -1.108349840867668e+04, 0.000000000000000e+00, -1.404560620237454e+04, 0.000000000000000e+00,
  -1.704875056093422e+04, 0.000000000000000e+00, -2.007948592176624e+04, 0.000000000000000e+00, -2.624146365257912e+05,
  0.000000000000000e+00, -2.619216402774265e+04, 0.000000000000000e+00, -2.926446089914285e+04, 0.000000000000000e+00,
  -3.234334611966101e+04, 0.000000000000000e+00, -3.542686654160940e+04, 0.000000000000000e+00, -3.851360938536154e+04,
  0.000000000000000e+00, -4.160254163637140e+04, 0.000000000000000e+00, -4.469289954118110e+04, 0.000000000000000e+00,
  -4.778411842524911e+04, 0.000000000000000e+00, -5.087577340656213e+04, 0.000000000000000e+00, -5.396755893046919e+04,
  0.000000000000000e+00, -5.705922730671355e+04, 0.000000000000000e+00, -6.015060788841274e+04, 0.000000000000000e+00,
  -6.324156638465029e+04, 0.000000000000000e+00, -6.633201159853510e+04, 0.000000000000000e+00, -6.942186879319654e+04,
  0.000000000000000e+00, 0.000000000000000e+00, -1.712905067652055e+03, 0.000000000000000e+00, -4.315578064443836e+03,
  0.000000000000000e+00, -7.204439940935847e+03, 0.000000000000000e+00, -1.024318449115097e+04, 0.000000000000000e+00,
  -1.337173102193114e+04, 0.000000000000000e+00, -1.655792346161379e+04, 0.000000000000000e+00, -1.978273523200483e+04,
  0.000000000000000e+00, -2.303417126438240e+04, 0.000000000000000e+00, -3.103492450272629e+05, 0.000000000000000e+00,
  -2.958782723645080e+04, 0.000000000000000e+00, -3.288090506877683e+04, 0.000000000000000e+00, -3.618088167443570e+04,
  0.000000000000000e+00, -3.948581526972262e+04, 0.000000000000000e+00, -4.279427809918152e+04, 0.000000000000000e+00,
  -4.610521147974072e+04, 0.000000000000000e+00, -4.941782224641141e+04, 0.000000000000000e+00, -5.273150955237660e+04,
  0.000000000000000e+00, -5.604581666765190e+04, 0.000000000000000e+00, -5.936041915863466e+04, 0.000000000000000e+00,
  -6.267503485593518e+04, 0.000000000000000e+00, -6.598947204350910e+04, 0.000000000000000e+00, -6.930358325849233e+04,
  0.000000000000000e+00, -7.261724620819326e+04, 0.000000000000000e+00, -7.593038645687095e+04, -6.150342792154646e+02,
  0.000000000000000e+00, -3.108161365902908e+03, 0.000000000000000e+00, -6.052884380854417e+03, 0.000000000000000e+00,
  -9.211482516356036e+03, 0.000000000000000e+00, -1.249403222670351e+04, 0.000000000000000e+00, -1.585501716657124e+04,
  0.000000000000000e+00, -1.926825374289000e+04, 0.000000000000000e+00, -2.271750262761901e+04, 0.000000000000000e+00,
  -2.619216402774265e+04, 0.000000000000000e+00, -3.637495448630142e+05, 0.000000000000000e+00, -3.319118897037260e+04,
  0.000000000000000e+00, -3.670698571953629e+04, 0.000000000000000e+00, -4.022986543064952e+04, 0.000000000000000e+00,
  -4.375792396784104e+04, 0.000000000000000e+00, -4.728974169170205e+04, 0.000000000000000e+00, -5.082425085135181e+04,
  0.000000000000000e+00, -5.436064037890825e+04, 0.000000000000000e+00, -5.789828887929389e+04, 0.000000000000000e+00,
  -6.143672221374641e+04, 0.000000000000000e+00, -6.497557049896402e+04, 0.000000000000000e+00, -6.851456950582970e+04,
  0.000000000000000e+00, -7.205348097015335e+04, 0.000000000000000e+00, -7.559214447773107e+04, 0.000000000000000e+00,
  -7.913043619237986e+04, 0.000000000000000e+00, 0.000000000000000e+00, -1.906507802644545e+03, 0.000000000000000e+00,
  -4.812484119847105e+03, 0.000000000000000e+00, -8.048028238161050e+03, 0.000000000000000e+00, -1.146034647150601e+04,
  0.000000000000000e+00, -1.498103270150460e+04, 0.000000000000000e+00, -1.857285239980273e+04, 0.000000000000000e+00,
  -2.221338355931084e+04, 0.000000000000000e+00, -2.588825260594868e+04, 0.000000000000000e+00, -2.958782723645080e+04,
  0.000000000000000e+00, -4.229418131073258e+05, 0.000000000000000e+00, -3.703635975673005e+04, 0.000000000000000e+00,
  -4.077715415715478e+04, 0.000000000000000e+00, -4.452529823146044e+04, 0.000000000000000e+00, -4.827890656232809e+04,
  0.000000000000000e+00, -5.203655453947624e+04, 0.000000000000000e+00, -5.579715560416423e+04, 0.000000000000000e+00,
  -5.955987546508196e+04, 0.000000000000000e+00, -6.332407160702428e+04, 0.000000000000000e+00, -6.708923988695626e+04,
  0.000000000000000e+00, -7.085498364644295e+04, 0.000000000000000e+00, -7.462102305104907e+04, 0.000000000000000e+00,
  -7.838709978335358e+04, 0.000000000000000e+00, -8.215302821529029e+04, 0.000000000000000e+00, -8.591866661511076e+04,
  -6.803121949128343e+02, 0.000000000000000e+00, -3.443671281577793e+03, 0.000000000000000e+00, -6.717258033018387e+03,
  0.000000000000000e+00, -1.023782368359479e+04, 0.000000000000000e+00, -1.390457634887765e+04, 0.000000000000000e+00,
  -1.766574600903898e+04, 0.000000000000000e+00, -2.149108363413323e+04, 0.000000000000000e+00, -2.536153099287273e+04,
  0.000000000000000e+00, -2.926446089914285e+04, 0.000000000000000e+00, -3.319118897037260e+04, 0.000000000000000e+00,
  -4.881883799115219e+05, 0.000000000000000e+00, -4.109317979367616e+04, 0.000000000000000e+00, -4.506073689335497e+04,
  0.000000000000000e+00, -4.903580129439737e+04, 0.000000000000000e+00, -5.301652971101602e+04, 0.000000000000000e+00,
  -5.700150796801807e+04, 0.000000000000000e+00, -6.098964462955694e+04, 0.000000000000000e+00, -6.498009835740041e+04,
  0.000000000000000e+00, -6.897220478652872e+04, 0.000000000000000e+00, -7.296544626820096e+04, 0.000000000000000e+00,
  -7.695941442901237e+04, 0.000000000000000e+00, -8.095377983256364e+04, 0.000000000000000e+00, -8.494831112680239e+04,
  0.000000000000000e+00, -8.894278200496297e+04, 0.000000000000000e+00, 0.000000000000000e+00, -2.099969957849160e+03,
  0.000000000000000e+00, -5.308906194856751e+03, 0.000000000000000e+00, -8.890936283542085e+03, 0.000000000000000e+00,
  -1.267704254574228e+04, 0.000000000000000e+00, -1.659064676188475e+04, 0.000000000000000e+00, -2.058949154564191e+04,
  0.000000000000000e+00, -2.464777235380881e+04, 0.000000000000000e+00, -2.874870938689374e+04, 0.000000000000000e+00,
  -3.288090506877683e+04, 0.000000000000000e+00, -3.703635975673005e+04, 0.000000000000000e+00, -5.598153306945488e+05,
  0.000000000000000e+00, -4.539555792210088e+04, 0.000000000000000e+00, -4.959192459877577e+04, 0.000000000000000e+00,
  -5.379603180278060e+04, 0.000000000000000e+00, -5.800605535560208e+04, 0.000000000000000e+00, -6.222058171382692e+04,
  0.000000000000000e+00, -6.643851461242813e+04, 0.000000000000000e+00, -7.065899036930337e+04, 0.000000000000000e+00,
  -7.488132645382313e+04, 0.000000000000000e+00, -7.910498512816726e+04, 0.000000000000000e+00, -8.332953839359432e+04,
  0.000000000000000e+00, -8.755462716230177e+04, 0.000000000000000e+00, -9.178001700657923e+04, 0.000000000000000e+00,
  -9.600545783505045e+04, -7.455617826369836e+02, 0.000000000000000e+00, -3.778888613051527e+03, 0.000000000000000e+00,
  -7.381020454526882e+03, 0.000000000000000e+00, -1.126347034788494e+04, 0.000000000000000e+00, -1.531478093423432e+04,
  0.000000000000000e+00, -1.947701802402329e+04, 0.000000000000000e+00, -2.371591060733285e+04, 0.000000000000000e+00,
  -2.800958061992193e+04, 0.000000000000000e+00, -3.234334611966101e+04, 0.000000000000000e+00, -3.670698571953629e+04,
  0.000000000000000e+00, -4.109317979367616e+04, 0.000000000000000e+00, -6.380847923629555e+05, 0.000000000000000e+00,
  -4.991317760308241e+04, 0.000000000000000e+00, -5.433997278241193e+04, 0.000000000000000e+00, -5.877464780337032e+04,
  0.000000000000000e+00, -6.321541825084870e+04, 0.000000000000000e+00, -6.766088330185179e+04, 0.000000000000000e+00,
  -7.210994600574786e+04, 0.000000000000000e+00, -7.656173842347811e+04, 0.000000000000000e+00, -8.101556934461255e+04,
  0.000000000000000e+00, -8.547087777129757e+04, 0.000000000000000e+00, -8.992722844202122e+04, 0.000000000000000e+00,
  -9.438425218976723e+04, 0.000000000000000e+00, -9.884166498032387e+04, 0.000000000000000e+00, 0.000000000000000e+00,
  -2.293316673891452e+03, 0.000000000000000e+00, -5.804906348815684e+03, 0.000000000000000e+00, -9.733182233855274e+03,
  0.000000000000000e+00, -1.389313330238267e+04, 0.000000000000000e+00, -1.820013706400732e+04, 0.000000000000000e+00,
  -2.260698818343645e+04, 0.000000000000000e+00, -2.708452528261843e+04, 0.000000000000000e+00, -3.161355207786552e+04,
  0.000000000000000e+00, -3.618088167443570e+04, 0.000000000000000e+00, -4.077715415715478e+04, 0.000000000000000e+00,
  -4.539555792210088e+04, 0.000000000000000e+00, -7.233226903542761e+05, 0.000000000000000e+00, -5.467977892397175e+04,
  0.000000000000000e+00, -5.933886239174958e+04, 0.000000000000000e+00, -6.400602717907717e+04, 0.000000000000000e+00,
  -6.867951648979403e+04, 0.000000000000000e+00, -7.335793086217259e+04, 0.000000000000000e+00, -7.804017123414487e+04,
  0.000000000000000e+00, -8.272535898703843e+04, 0.000000000000000e+00, -8.741278005124502e+04, 0.000000000000000e+00,
  -9.210186161978675e+04, 0.000000000000000e+00, -9.679215025591958e+04, 0.000000000000000e+00, -1.014832693794901e+05,
  0.000000000000000e+00, -1.061748933336833e+05, -8.107886993610502e+02, 0.000000000000000e+00, -4.113858728504948e+03,
  0.000000000000000e+00, -8.044231668697851e+03, 0.000000000000000e+00, -1.228841447024543e+04, 0.000000000000000e+00,
  -1.672444964297231e+04, 0.000000000000000e+00, -2.128834434914897e+04, 0.000000000000000e+00, -2.594184174914374e+04,
  0.000000000000000e+00, -3.066026488679739e+04, 0.000000000000000e+00, -3.542686654160940e+04, 0.000000000000000e+00,
  -4.022986543064952e+04, 0.000000000000000e+00, -4.506073689335497e+04, 0.000000000000000e+00, -4.991317760308241e+04,
  0.000000000000000e+00, -8.157909850802867e+05, 0.000000000000000e+00, -5.966491315564311e+04, 0.000000000000000e+00,
  -6.455777960407246e+04, 0.000000000000000e+00, -6.945885266969733e+04, 0.000000000000000e+00, -7.436640646165767e+04,
  0.000000000000000e+00, -7.927906485939310e+04, 0.000000000000000e+00, -8.419572896217088e+04, 0.000000000000000e+00,
  -8.911551453110010e+04, 0.000000000000000e+00, -9.403770040977820e+04, 0.000000000000000e+00, -9.896171132859382e+04,
  0.000000000000000e+00, -1.038870673979919e+05, 0.000000000000000e+00, -1.088133851458092e+05, 0.000000000000000e+00,
  0.000000000000000e+00, -2.486567997506928e+03, 0.000000000000000e+00, -6.300536986920059e+03, 0.000000000000000e+00,
  -1.057480512673283e+04, 0.000000000000000e+00, -1.510854699002579e+04, 0.000000000000000e+00, -1.980923929204588e+04,
  0.000000000000000e+00, -2.462478625987730e+04, 0.000000000000000e+00, -2.952269358841811e+04, 0.000000000000000e+00,
  -3.448136609182136e+04, 0.000000000000000e+00, -3.948581526972262e+04, 0.000000000000000e+00, -4.452529823146044e+04,
  0.000000000000000e+00, -4.959192459877577e+04, 0.000000000000000e+00, -5.467977892397175e+04, 0.000000000000000e+00,
  -9.158154670744325e+05, 0.000000000000000e+00, -6.490217998399524e+04, 0.000000000000000e+00, -7.003053231308585e+04,
  0.000000000000000e+00, -7.516727239760374e+04, 0.000000000000000e+00, -8.031069449321255e+04, 0.000000000000000e+00,
  -8.545943288118076e+04, 0.000000000000000e+00, -9.061239058842056e+04, 0.000000000000000e+00, -9.576866682771713e+04,
  0.000000000000000e+00, -1.009275390174100e+05, 0.000000000000000e+00, -1.060884133930459e+05, 0.000000000000000e+00,
  -1.112507979164287e+05, 0.000000000000000e+00, -1.164142867761275e+05, -8.759971625273406e+02, 0.000000000000000e+00,
  -4.448618762383861e+03, 0.000000000000000e+00, -8.706945752089758e+03, 0.000000000000000e+00, -1.331267141125037e+04,
  0.000000000000000e+00, -1.813348257048470e+04, 0.000000000000000e+00, -2.309940740131831e+04, 0.000000000000000e+00,
  -2.816826899311096e+04, 0.000000000000000e+00, -3.331260690566686e+04, 0.000000000000000e+00, -3.851360938536154e+04,
  0.000000000000000e+00, -4.375792396784104e+04, 0.000000000000000e+00, -4.903580129439737e+04, 0.000000000000000e+00,
  -5.433997278241193e+04, 0.000000000000000e+00, -5.966491315564311e+04, 0.000000000000000e+00, -1.023657957888224e+06,
  0.000000000000000e+00, -7.036101740836033e+04, 0.000000000000000e+00, -7.572624705117718e+04, 0.000000000000000e+00,
  -8.109997032159523e+04, 0.000000000000000e+00, -8.648052498854331e+04, 0.000000000000000e+00, -9.186655318185870e+04,
  0.000000000000000e+00, -9.725696932020152e+04, 0.000000000000000e+00, -1.026508714598336e+05, 0.000000000000000e+00,
  -1.080475302131853e+05, 0.000000000000000e+00, -1.134463431692979e+05, 0.000000000000000e+00, -1.188468094045018e+05,
  0.000000000000000e+00, 0.000000000000000e+00, -2.679738577839147e+03, 0.000000000000000e+00, -6.795843707256664e+03,
  0.000000000000000e+00, -1.141584891033131e+04, 0.000000000000000e+00, -1.632327568583192e+04, 0.000000000000000e+00,
  -2.141779889431290e+04, 0.000000000000000e+00, -2.664250853429418e+04, 0.000000000000000e+00, -3.196161317596226e+04,
  0.000000000000000e+00, -3.735113609091195e+04, 0.000000000000000e+00, -4.279427809918152e+04, 0.000000000000000e+00,
  -4.827890656232809e+04, 0.000000000000000e+00, -5.379603180278060e+04, 0.000000000000000e+00, -5.933886239174958e+04,
  0.000000000000000e+00, -6.490217998399524e+04, 0.000000000000000e+00, -1.139644132916086e+06, 0.000000000000000e+00,
  -7.607490792562709e+04, 0.000000000000000e+00, -8.167858130822929e+04, 0.000000000000000e+00, -8.729091226637966e+04,
  0.000000000000000e+00, -9.291025221224020e+04, 0.000000000000000e+00, -9.853526627675911e+04, 0.000000000000000e+00,
  -1.041648595301516e+05, 0.000000000000000e+00, -1.097981309737959e+05, 0.000000000000000e+00, -1.154343408522441e+05,
  0.000000000000000e+00, -1.210728744749231e+05, 0.000000000000000e+00, -1.267132189422982e+05, -9.411903851588594e+02,
  0.000000000000000e+00, -4.783197432539635e+03, 0.000000000000000e+00, -9.369216356199058e+03, 0.000000000000000e+00,
  -1.433627404487471e+04, 0.000000000000000e+00, -1.954181899378723e+04, 0.000000000000000e+00, -2.491001115513829e+04,
  0.000000000000000e+00, -3.039478142617566e+04, 0.000000000000000e+00, -3.596591041251293e+04, 0.000000000000000e+00,
  -4.160254163637140e+04, 0.000000000000000e+00, -4.728974169170205e+04, 0.000000000000000e+00, -5.301652971101602e+04,
  0.000000000000000e+00, -5.877464780337032e+04, 0.000000000000000e+00, -6.455777960407246e+04, 0.000000000000000e+00,
  -7.036101740836033e+04, 0.000000000000000e+00, -1.264035695029677e+06, 0.000000000000000e+00, -8.201318900816797e+04,
  0.000000000000000e+00, -8.785660756516767e+04, 0.000000000000000e+00, -9.370877897793161e+04, 0.000000000000000e+00,
  -9.956809208153395e+04, 0.000000000000000e+00, -1.054332277567653e+05, 0.000000000000000e+00, -1.113030968206700e+05,
  0.000000000000000e+00, -1.171768036865132e+05, 0.000000000000000e+00, -1.230536017599951e+05, 0.000000000000000e+00,
  -1.289328670027522e+05, 0.000000000000000e+00, 0.000000000000000e+00, -2.872841567736994e+03, 0.000000000000000e+00,
  -7.290865619606987e+03, 0.000000000000000e+00, -1.225636298288041e+04, 0.000000000000000e+00, -1.753732390748895e+04,
  0.000000000000000e+00, -2.302573220300199e+04, 0.000000000000000e+00, -2.865991018748078e+04, 0.000000000000000e+00,
  -3.440082799216398e+04, 0.000000000000000e+00, -4.022212463470113e+04, 0.000000000000000e+00, -4.610521147974072e+04,
  0.000000000000000e+00, -5.203655453947624e+04, 0.000000000000000e+00, -5.800605535560208e+04, 0.000000000000000e+00,
  -6.400602717907717e+04, 0.000000000000000e+00, -7.003053231308585e+04, 0.000000000000000e+00, -7.607490792562709e+04,
  0.000000000000000e+00, -1.397158220819778e+06, 0.000000000000000e+00, -8.820924194815052e+04, 0.000000000000000e+00,
  -9.429385930738888e+04, 0.000000000000000e+00, -1.003873696352103e+05, 0.000000000000000e+00, -1.064881951742419e+05,
  0.000000000000000e+00, -1.125950168868675e+05, 0.000000000000000e+00, -1.187067546750715e+05, 0.000000000000000e+00,
  -1.248225091958411e+05, 0.000000000000000e+00, -1.309415265540089e+05, 0.000000000000000e+00, -1.370631749839225e+05,
  -1.006370862428824e+03, 0.000000000000000e+00, -5.117620385308717e+03, 0.000000000000000e+00, -1.003108929665355e+04,
  0.000000000000000e+00, -1.535926136292369e+04, 0.000000000000000e+00, -2.094945732202523e+04, 0.000000000000000e+00,
  -2.672003474298904e+04, 0.000000000000000e+00, -3.262109138599801e+04, 0.000000000000000e+00, -3.861967581614530e+04,
  0.000000000000000e+00, -4.469289954118110e+04, 0.000000000000000e+00, -5.082425085135181e+04, 0.000000000000000e+00,
  -5.700150796801807e+04, 0.000000000000000e+00, -6.321541825084870e+04, 0.000000000000000e+00, -6.945885266969733e+04,
  0.000000000000000e+00, -7.572624705117718e+04, 0.000000000000000e+00, -8.201318900816797e+04, 0.000000000000000e+00,
  -1.539273309565493e+06, 0.000000000000000e+00, -9.463231974434837e+04, 0.000000000000000e+00, -1.009593549544935e+05,
  0.000000000000000e+00, -1.072953743383482e+05, 0.000000000000000e+00, -1.136388182304298e+05, 0.000000000000000e+00,
  -1.199884019106897e+05, 0.000000000000000e+00, -1.263430455243320e+05, 0.000000000000000e+00, -1.327018495825507e+05,
  0.000000000000000e+00, -1.390640720849373e+05, 0.000000000000000e+00, 0.000000000000000e+00, -3.065885822368984e+03,
  0.000000000000000e+00, -7.785636442245068e+03, 0.000000000000000e+00, -1.309638705352119e+04, 0.000000000000000e+00,
  -1.875071407578987e+04, 0.000000000000000e+00, -2.463297826597230e+04, 0.000000000000000e+00, -3.067683003429458e+04,
  0.000000000000000e+00, -3.684000511242408e+04, 0.000000000000000e+00, -4.309378901743732e+04, 0.000000000000000e+00,
  -4.941782224641141e+04, 0.000000000000000e+00, -5.579715560416423e+04, 0.000000000000000e+00, -6.222058171382692e+04,
  0.000000000000000e+00, -6.867951648979403e+04, 0.000000000000000e+00, -7.516727239760374e+04, 0.000000000000000e+00,
  -8.167858130822929e+04, 0.000000000000000e+00, -8.820924194815052e+04, 0.000000000000000e+00, -1.690706450973286e+06,
  0.000000000000000e+00, -1.013157111470873e+05, 0.000000000000000e+00, -1.078865182130349e+05, 0.000000000000000e+00,
  -1.144664435814330e+05, 0.000000000000000e+00, -1.210539433299121e+05, 0.000000000000000e+00, -1.276477431001658e+05,
  0.000000000000000e+00, -1.342467723193540e+05, 0.000000000000000e+00, -1.408501353787975e+05, 0.000000000000000e+00,
  -1.474570753693506e+05, -1.071540565398104e+03, 0.000000000000000e+00, -5.451908258204180e+03, 0.000000000000000e+00,
  -1.069260622725354e+04, 0.000000000000000e+00, -1.638167347812988e+04, 0.000000000000000e+00, -2.235640155059753e+04,
  0.000000000000000e+00, -2.852941452646049e+04, 0.000000000000000e+00, -3.484700783249143e+04, 0.000000000000000e+00,
  -4.127355149244117e+04, 0.000000000000000e+00, -4.778411842524911e+04, 0.000000000000000e+00, -5.436064037890825e+04,
  0.000000000000000e+00, -6.098964462955694e+04, 0.000000000000000e+00, -6.766088330185179e+04, 0.000000000000000e+00,
  -7.436640646165767e+04, 0.000000000000000e+00, -8.109997032159523e+04, 0.000000000000000e+00, -8.785660756516767e+04,
  0.000000000000000e+00, -9.463231974434837e+04, 0.000000000000000e+00, -1.851719157342708e+06, 0.000000000000000e+00,
  -1.082285928569243e+05, 0.000000000000000e+00, -1.150443254313490e+05, 0.000000000000000e+00, -1.218692497540516e+05,
  0.000000000000000e+00, -1.287018613950925e+05, 0.000000000000000e+00, -1.355408948991081e+05, 0.000000000000000e+00,
  -1.423852948747153e+05, 0.000000000000000e+00, -1.492341706069093e+05, 0.000000000000000e+00, 0.000000000000000e+00,
  -3.258879984613314e+03, 0.000000000000000e+00, -8.280185354358528e+03, 0.000000000000000e+00, -1.393596630073454e+04,
  0.000000000000000e+00, -1.996347869006161e+04, 0.000000000000000e+00, -2.623953905598289e+04, 0.000000000000000e+00,
  -3.269316790123957e+04, 0.000000000000000e+00, -3.927891667675359e+04, 0.000000000000000e+00, -4.596574782348224e+04,
  0.000000000000000e+00, -5.273150955237660e+04, 0.000000000000000e+00, -5.955987546508196e+04, 0.000000000000000e+00,
  -6.643851461242813e+04, 0.000000000000000e+00, -7.335793086217259e+04, 0.000000000000000e+00, -8.031069449321255e+04,
  0.000000000000000e+00, -8.729091226637966e+04, 0.000000000000000e+00, -9.429385930738888e+04, 0.000000000000000e+00,
  -1.013157111470873e+05, 0.000000000000000e+00, -2.022636842695316e+06, 0.000000000000000e+00, -1.154041829668304e+05,
  0.000000000000000e+00, -1.224661036914984e+05, 0.000000000000000e+00, -1.295373333194179e+05, 0.000000000000000e+00,
  -1.366163901566036e+05, 0.000000000000000e+00, -1.437020296197055e+05, 0.000000000000000e+00, -1.507931889454310e+05,
  0.000000000000000e+00, -1.578889718172165e+05, -1.136701073987475e+03, 0.000000000000000e+00, -5.786076011711925e+03,
  0.000000000000000e+00, -1.135380234444897e+04, 0.000000000000000e+00, -1.740355714564152e+04, 0.000000000000000e+00,
  -2.376266941878928e+04, 0.000000000000000e+00, -3.033809183099653e+04, 0.000000000000000e+00, -3.707240069370350e+04,
  0.000000000000000e+00, -4.392727276020605e+04, 0.000000000000000e+00, -5.087577340656213e+04, 0.000000000000000e+00,
  -5.789828887929389e+04, 0.000000000000000e+00, -6.498009835740041e+04, 0.000000000000000e+00, -7.210994600574786e+04,
  0.000000000000000e+00, -7.927906485939310e+04, 0.000000000000000e+00, -8.648052498854331e+04, 0.000000000000000e+00,
  -9.370877897793161e+04, 0.000000000000000e+00, -1.009593549544935e+05, 0.000000000000000e+00, -1.082285928569243e+05,
  0.000000000000000e+00, -2.203720936071291e+06, 0.000000000000000e+00, -1.228115872472657e+05, 0.000000000000000e+00,
  -1.301207838097130e+05, 0.000000000000000e+00, -1.374393668921790e+05, 0.000000000000000e+00, -1.447658683149671e+05,
  0.000000000000000e+00, -1.520990686318470e+05, 0.000000000000000e+00, -1.594379221695777e+05, 0.000000000000000e+00,
  0.000000000000000e+00, -3.451830367885173e+03, 0.000000000000000e+00, -8.774536985201288e+03, 0.000000000000000e+00,
  -1.477514032106358e+04, 0.000000000000000e+00, -2.117565016302377e+04, 0.000000000000000e+00, -2.784541112097300e+04,
  0.000000000000000e+00, -3.470887072242604e+04, 0.000000000000000e+00, -4.171740431337220e+04, 0.000000000000000e+00,
  -4.883769700288491e+04, 0.000000000000000e+00, -5.604581666765190e+04, 0.000000000000000e+00, -6.332407160702428e+04,
  0.000000000000000e+00, -7.065899036930337e+04, 0.000000000000000e+00, -7.804017123414487e+04, 0.000000000000000e+00,
  -8.545943288118076e+04, 0.000000000000000e+00, -9.291025221224020e+04, 0.000000000000000e+00, -1.003873696352103e+05,
  0.000000000000000e+00, -1.078865182130349e+05, 0.000000000000000e+00, -1.154041829668304e+05, 0.000000000000000e+00,
  -2.395296791846300e+06, 0.000000000000000e+00, -1.304839475653954e+05, 0.000000000000000e+00, -1.380416178029519e+05,
  0.000000000000000e+00, -1.456087738091706e+05, 0.000000000000000e+00, -1.531839832281688e+05, 0.000000000000000e+00,
  -1.607660232672886e+05, 0.000000000000000e+00, -1.683538603102943e+05, -1.201853672227242e+03, 0.000000000000000e+00,
  -6.120139641823457e+03, 0.000000000000000e+00, -1.201470818596153e+04, 0.000000000000000e+00, -1.842494263862793e+04,
  0.000000000000000e+00, -2.516828985162449e+04, 0.000000000000000e+00, -3.214607770814979e+04, 0.000000000000000e+00,
  -3.929718873955939e+04, 0.000000000000000e+00, -4.658065548237089e+04, 0.000000000000000e+00, -5.396755893046919e+04,
  0.000000000000000e+00, -6.143672221374641e+04, 0.000000000000000e+00, -6.897220478652872e+04, 0.000000000000000e+00,
  -7.656173842347811e+04, 0.000000000000000e+00, -8.419572896217088e+04, 0.000000000000000e+00, -9.186655318185870e+04,
  0.000000000000000e+00, -9.956809208153395e+04, 0.000000000000000e+00, -1.072953743383482e+05, 0.000000000000000e+00,
  -1.150443254313490e+05, 0.000000000000000e+00, -1.228115872472657e+05, 0.000000000000000e+00, -2.597625762724247e+06,
  0.000000000000000e+00, -1.383903180965494e+05, 0.000000000000000e+00, -1.461974832277765e+05, 0.000000000000000e+00,
  -1.540142073522489e+05, 0.000000000000000e+00, -1.618390698236272e+05, 0.000000000000000e+00, -1.696708789181237e+05,
  0.000000000000000e+00, 0.000000000000000e+00, -3.644742275450894e+03, 0.000000000000000e+00, -9.268710549823660e+03,
  0.000000000000000e+00, -1.561394176606279e+04, 0.000000000000000e+00, -2.238726351686981e+04, 0.000000000000000e+00,
  -2.945060909629598e+04, 0.000000000000000e+00, -3.672387439603401e+04, 0.000000000000000e+00, -4.415535676459959e+04,
  0.000000000000000e+00, -5.170942695435766e+04, 0.000000000000000e+00, -5.936041915863466e+04, 0.000000000000000e+00,
  -6.708923988695626e+04, 0.000000000000000e+00, -7.488132645382313e+04, 0.000000000000000e+00, -8.272535898703843e+04,
  0.000000000000000e+00, -9.061239058842056e+04, 0.000000000000000e+00, -9.853526627675911e+04, 0.000000000000000e+00,
  -1.064881951742419e+05, 0.000000000000000e+00, -1.144664435814330e+05, 0.000000000000000e+00, -1.224661036914984e+05,
  0.000000000000000e+00, -1.304839475653954e+05, 0.000000000000000e+00, -2.811033144531446e+06, 0.000000000000000e+00,
  -1.465637755084384e+05, 0.000000000000000e+00, -1.546215714436042e+05, 0.000000000000000e+00, -1.626890125485485e+05,
  0.000000000000000e+00, -1.707647140674428e+05, 0.000000000000000e+00, -1.788474895477776e+05, -1.266999411449095e+03,
  0.000000000000000e+00, -6.454110686901437e+03, 0.000000000000000e+00, -1.267535271733391e+04, 0.000000000000000e+00,
  -1.944587313277898e+04, 0.000000000000000e+00, -2.657329120666107e+04, 0.000000000000000e+00, -3.395336555015762e+04,
  0.000000000000000e+00, -4.152133388419947e+04, 0.000000000000000e+00, -4.923356999207695e+04, 0.000000000000000e+00,
  -5.705922730671355e+04, 0.000000000000000e+00, -6.497557049896402e+04, 0.000000000000000e+00, -7.296544626820096e+04,
  0.000000000000000e+00, -8.101556934461255e+04, 0.000000000000000e+00, -8.911551453110010e+04, 0.000000000000000e+00,
  -9.725696932020152e+04, 0.000000000000000e+00, -1.054332277567653e+05, 0.000000000000000e+00, -1.136388182304298e+05,
  0.000000000000000e+00, -1.218692497540516e+05, 0.000000000000000e+00, -1.301207838097130e+05, 0.000000000000000e+00,
  -1.383903180965494e+05, 0.000000000000000e+00, -3.035780236560512e+06, 0.000000000000000e+00, -1.549733261253163e+05,
  0.000000000000000e+00, -1.632827202351223e+05, 0.000000000000000e+00, -1.716018242726882e+05, 0.000000000000000e+00,
  -1.799292745511657e+05, 0.000000000000000e+00, 0.000000000000000e+00, -3.837620686586279e+03, 0.000000000000000e+00,
  -9.762725985126383e+03, 0.000000000000000e+00, -1.645240485431484e+04, 0.000000000000000e+00, -2.359835989651236e+04,
  0.000000000000000e+00, -3.105515224699034e+04, 0.000000000000000e+00, -3.873819661472266e+04, 0.000000000000000e+00,
  -4.659270496871221e+04, 0.000000000000000e+00, -5.458078395316319e+04, 0.000000000000000e+00, -6.267503485593518e+04,
  0.000000000000000e+00, -7.085498364644295e+04, 0.000000000000000e+00, -7.910498512816726e+04, 0.000000000000000e+00,
  -8.741278005124502e+04, 0.000000000000000e+00, -9.576866682771713e+04, 0.000000000000000e+00, -1.041648595301516e+05,
  0.000000000000000e+00, -1.125950168868675e+05, 0.000000000000000e+00, -1.210539433299121e+05, 0.000000000000000e+00,
  -1.295373333194179e+05, 0.000000000000000e+00, -1.380416178029519e+05, 0.000000000000000e+00, -1.465637755084384e+05,
  0.000000000000000e+00, -3.272192268456585e+06, 0.000000000000000e+00, -1.636519838504277e+05, 0.000000000000000e+00,
  -1.722140479312287e+05, 0.000000000000000e+00, -1.807859047113839e+05, 0.000000000000000e+00, -1.893662256332006e+05,
  -1.332139160330905e+03, 0.000000000000000e+00, -6.787998439043877e+03, 0.000000000000000e+00, -1.333576218715648e+04,
  0.000000000000000e+00, -2.046638393962805e+04, 0.000000000000000e+00, -2.797770788827053e+04, 0.000000000000000e+00,
  -3.575996760516313e+04, 0.000000000000000e+00, -4.374476465282483e+04, 0.000000000000000e+00, -5.188592439452296e+04,
  0.000000000000000e+00, -6.015060788841274e+04, 0.000000000000000e+00, -6.851456950582970e+04, 0.000000000000000e+00,
  -7.695941442901237e+04, 0.000000000000000e+00, -8.547087777129757e+04, 0.000000000000000e+00, -9.403770040977820e+04,
  0.000000000000000e+00, -1.026508714598336e+05, 0.000000000000000e+00, -1.113030968206700e+05, 0.000000000000000e+00,
  -1.199884019106897e+05, 0.000000000000000e+00, -1.287018613950925e+05, 0.000000000000000e+00, -1.374393668921790e+05,
  0.000000000000000e+00, -1.461974832277765e+05, 0.000000000000000e+00, -1.549733261253163e+05, 0.000000000000000e+00,
  -3.520530486266942e+06, 0.000000000000000e+00, -1.725687011866444e+05, 0.000000000000000e+00, -1.813843676848297e+05,
  0.000000000000000e+00, -1.902098840589359e+05, 0.000000000000000e+00, 0.000000000000000e+00, -4.030469242118391e+03,
  0.000000000000000e+00, -1.025659846127254e+04, 0.000000000000000e+00, -1.729055685911823e+04, 0.000000000000000e+00,
  -2.480895994525463e+04, 0.000000000000000e+00, -3.265906906650026e+04, 0.000000000000000e+00, -4.075182608300554e+04,
  0.000000000000000e+00, -4.902941583009261e+04, 0.000000000000000e+00, -5.745165428250096e+04, 0.000000000000000e+00,
  -6.598947204350910e+04, 0.000000000000000e+00, -7.462102305104907e+04, 0.000000000000000e+00, -8.332953839359432e+04,
  0.000000000000000e+00, -9.210186161978675e+04, 0.000000000000000e+00, -1.009275390174100e+05, 0.000000000000000e+00,
  -1.097981309737959e+05, 0.000000000000000e+00, -1.187067546750715e+05, 0.000000000000000e+00, -1.276477431001658e+05,
  0.000000000000000e+00, -1.366163901566036e+05, 0.000000000000000e+00, -1.456087738091706e+05, 0.000000000000000e+00,
  -1.546215714436042e+05, 0.000000000000000e+00, -1.636519838504277e+05, 0.000000000000000e+00, -3.781120076607686e+06,
  0.000000000000000e+00, -1.817564593731948e+05, 0.000000000000000e+00, -1.908267361852324e+05, 0.000000000000000e+00,
  -1.999069497420510e+05, -1.397273639967767e+03, 0.000000000000000e+00, -7.121813205198619e+03, 0.000000000000000e+00,
  -1.399595621755738e+04, 0.000000000000000e+00, -2.148650379292331e+04, 0.000000000000000e+00, -2.938156716047105e+04,
  0.000000000000000e+00, -3.756590081506492e+04, 0.000000000000000e+00, -4.596751016628496e+04, 0.000000000000000e+00,
  -5.453765658536724e+04, 0.000000000000000e+00, -6.324156638465029e+04, 0.000000000000000e+00, -7.205348097015335e+04,
  0.000000000000000e+00, -8.095377983256364e+04, 0.000000000000000e+00, -8.992722844202122e+04, 0.000000000000000e+00,
  -9.896171132859382e+04, 0.000000000000000e+00, -1.080475302131853e+05, 0.000000000000000e+00, -1.171768036865132e+05,
  0.000000000000000e+00, -1.263430455243320e+05, 0.000000000000000e+00, -1.355408948991081e+05, 0.000000000000000e+00,
  -1.447658683149671e+05, 0.000000000000000e+00, -1.540142073522489e+05, 0.000000000000000e+00, -1.632827202351223e+05,
  0.000000000000000e+00, -1.725687011866444e+05, 0.000000000000000e+00, -4.054222235926325e+06, 0.000000000000000e+00,
  -1.911841608988247e+05, 0.000000000000000e+00, -2.005099425724675e+05, 0.000000000000000e+00, 0.000000000000000e+00,
  -4.223291037012832e+03, 0.000000000000000e+00, -1.075034369139816e+04, 0.000000000000000e+00, -1.812841989874500e+04,
  0.000000000000000e+00, -2.601910769426565e+04, 0.000000000000000e+00, -3.426238446975283e+04, 0.000000000000000e+00,
  -4.276477248623956e+04, 0.000000000000000e+00, -5.146541270251458e+04, 0.000000000000000e+00, -6.032195726519755e+04,
  0.000000000000000e+00, -6.930358325849233e+04, 0.000000000000000e+00, -7.838709978335358e+04, 0.000000000000000e+00,
  -8.755462716230177e+04, 0.000000000000000e+00, -9.679215025591958e+04, 0.000000000000000e+00, -1.060884133930459e+05,
  0.000000000000000e+00, -1.154343408522441e+05, 0.000000000000000e+00, -1.248225091958411e+05, 0.000000000000000e+00,
  -1.342467723193540e+05, 0.000000000000000e+00, -1.437020296197055e+05, 0.000000000000000e+00, -1.531839832281688e+05,
  0.000000000000000e+00, -1.626890125485485e+05, 0.000000000000000e+00, -1.722140479312287e+05, 0.000000000000000e+00,
  -1.817564593731948e+05, 0.000000000000000e+00, -4.340162103443467e+06, 0.000000000000000e+00, -2.008847336309506e+05,
  0.000000000000000e+00, -2.104669723520248e+05, -1.462403456704839e+03, 0.000000000000000e+00, -7.455561197379555e+03,
  0.000000000000000e+00, -1.465595513085652e+04, 0.000000000000000e+00, -2.250626287401369e+04, 0.000000000000000e+00,
  -3.078491197016141e+04, 0.000000000000000e+00, -3.937119190166871e+04, 0.000000000000000e+00, -4.818955733501816e+04,
  0.000000000000000e+00, -5.718875255823218e+04, 0.000000000000000e+00, -6.633201159853510e+04, 0.000000000000000e+00,
  -7.559214447773107e+04, 0.000000000000000e+00, -8.494831112680239e+04, 0.000000000000000e+00, -9.438425218976723e+04,
  0.000000000000000e+00, -1.038870673979919e+05, 0.000000000000000e+00, -1.134463431692979e+05, 0.000000000000000e+00,
  -1.230536017599951e+05, 0.000000000000000e+00, -1.327018495825507e+05, 0.000000000000000e+00, -1.423852948747153e+05,
  0.000000000000000e+00, -1.520990686318470e+05, 0.000000000000000e+00, -1.618390698236272e+05, 0.000000000000000e+00,
  -1.716018242726882e+05, 0.000000000000000e+00, -1.813843676848297e+05, 0.000000000000000e+00, -1.911841608988247e+05,
  0.000000000000000e+00, -4.639200828003288e+06, 0.000000000000000e+00, -2.108270401106518e+05, 0.000000000000000e+00,
  0.000000000000000e+00, -4.416089645564901e+03, 0.000000000000000e+00, -1.124396941732769e+04, 0.000000000000000e+00,
  -1.896602166840895e+04, 0.000000000000000e+00, -2.722882647230378e+04, 0.000000000000000e+00, -3.586512709803659e+04,
  0.000000000000000e+00, -4.477704804260290e+04, 0.000000000000000e+00, -5.390073605170132e+04, 0.000000000000000e+00,
  -6.319164134422907e+04, 0.000000000000000e+00, -7.261724620819326e+04, 0.000000000000000e+00, -8.215302821529029e+04,
  0.000000000000000e+00, -9.178001700657923e+04, 0.000000000000000e+00, -1.014832693794901e+05, 0.000000000000000e+00,
  -1.112507979164287e+05, 0.000000000000000e+00, -1.210728744749231e+05, 0.000000000000000e+00, -1.309415265540089e+05,
  0.000000000000000e+00, -1.408501353787975e+05, 0.000000000000000e+00, -1.507931889454310e+05, 0.000000000000000e+00,
  -1.607660232672886e+05, 0.000000000000000e+00, -1.707647140674428e+05, 0.000000000000000e+00, -1.807859047113839e+05,
  0.000000000000000e+00, -1.908267361852324e+05, 0.000000000000000e+00, -2.008847336309506e+05, 0.000000000000000e+00,
  -4.951663510385709e+06, 0.000000000000000e+00, -2.210439689001755e+05, -1.527529121258250e+03, 0.000000000000000e+00,
  -7.789249104757998e+03, 0.000000000000000e+00, -1.531577253265984e+04, 0.000000000000000e+00, -2.352568895009513e+04,
  0.000000000000000e+00, -3.218775443098358e+04, 0.000000000000000e+00, -4.117586089622098e+04, 0.000000000000000e+00,
  -5.041091321403611e+04, 0.000000000000000e+00, -5.983911452224748e+04, 0.000000000000000e+00, -6.942186879319654e+04,
  0.000000000000000e+00, -7.913043619237986e+04, 0.000000000000000e+00, -8.894278200496297e+04, 0.000000000000000e+00,
  -9.884166498032387e+04, 0.000000000000000e+00, -1.088133851458092e+05, 0.000000000000000e+00, -1.188468094045018e+05,
  0.000000000000000e+00, -1.289328670027522e+05, 0.000000000000000e+00, -1.390640720849373e+05, 0.000000000000000e+00,
  -1.492341706069093e+05, 0.000000000000000e+00, -1.594379221695777e+05, 0.000000000000000e+00, -1.696708789181237e+05,
  0.000000000000000e+00, -1.799292745511657e+05, 0.000000000000000e+00, -1.902098840589359e+05, 0.000000000000000e+00,
  -2.005099425724675e+05, 0.000000000000000e+00, -2.108270401106518e+05, 0.000000000000000e+00, -5.277811258825032e+06,
  0.000000000000000e+00, 0.000000000000000e+00, -4.608866029704063e+03, 0.000000000000000e+00, -1.173749012291644e+04,
  0.000000000000000e+00, -1.980338156304536e+04, 0.000000000000000e+00, -2.843814955052699e+04, 0.000000000000000e+00,
  -3.746732406705313e+04, 0.000000000000000e+00, -4.678867343710619e+04, 0.000000000000000e+00, -5.633536375283465e+04,
  0.000000000000000e+00, -6.606069443269838e+04, 0.000000000000000e+00, -7.593038645687095e+04, 0.000000000000000e+00,
  -8.591866661511076e+04, 0.000000000000000e+00, -9.600545783505045e+04, 0.000000000000000e+00, -1.061748933336833e+05,
  0.000000000000000e+00, -1.164142867761275e+05, 0.000000000000000e+00, -1.267132189422982e+05, 0.000000000000000e+00,
  -1.370631749839225e+05, 0.000000000000000e+00, -1.474570753693506e+05, 0.000000000000000e+00, -1.578889718172165e+05,
  0.000000000000000e+00, -1.683538603102943e+05, 0.000000000000000e+00, -1.788474895477776e+05, 0.000000000000000e+00,
  -1.893662256332006e+05, 0.000000000000000e+00, -1.999069497420510e+05, 0.000000000000000e+00, -2.104669723520248e+05,
  0.000000000000000e+00, -2.210439689001755e+05, 0.000000000000000e+00, -5.617969136927102e+06,
};

#undef  Ac
#define Ac(i,j)  A48[ Nc*(j) + i ]

Real VolWaveDrag::dragCoefficient(Real Sref, Real L, const Vector &ak)
{
  Real cdw = 0.0;
  for (int j=0; j<Nc; ++j)
    for (int i=0; i<Nc; ++i)
      cdw += ak[i]*Ac(i,j)*ak[j];

  //  // debug
  //  Vector aAa(Nc);
  //  aAa = 0.0;
  //  for (int j=0; j<Nc; ++j)
  //    for (int i=0; i<Nc; ++i)
  //      aAa[i] += ak[i]*Ac(i,j)*ak[j];

  //  cout << "ak, ak .* A*ak" << endl;
  //  for (int i=0; i<Nc; ++i)
  //    cout << i+1 << " : " << ak[i] << ",  " << aAa[i] << endl;

  return -cdw / (2*M_PI*Sref*sq(L));
}

#undef Ac


